"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

TAGTYPE_MOUNTOPT_VERS=1
TAGTYPE_DETECTED=2
TAGTYPE_USER=3

TAG_DELEG_ENABLED='deleg'
TAG_PNFS_ENABLED='pnfs'

TAG_DESC = {
    'v2':                    TAGTYPE_MOUNTOPT_VERS,
    'v3':                    TAGTYPE_MOUNTOPT_VERS,
    'v4':                    TAGTYPE_MOUNTOPT_VERS,
    'v4.1':                  TAGTYPE_MOUNTOPT_VERS,

    TAG_DELEG_ENABLED: TAGTYPE_DETECTED,
    TAG_PNFS_ENABLED:  TAGTYPE_DETECTED,
}

def validate_tags(tags):
    r = []
    for t in tags:
        if t in TAG_DESC:
            r.append('%s is a reserved tag' % (t))
    return r

def nfstags(*tags):
    tags = list(set(tags))
    tags.sort()
    return tags

def make_nfsopt(mountopt, tags):
    nfsopt = mountopt

    if tags:
        nfsopt += ',%s' % ','.join(tags)

    return nfsopt

def nfsopt_normalize_version(nfsopts, old_syntax=False):
    if isinstance(nfsopts, (list, tuple)):
        ret = []
        for n in nfsopts:
            ret.append(_nfsopt_normalize_version(n, old_syntax=old_syntax))
        return ret
    else:
        return _nfsopt_normalize_version(nfsopts, old_syntax=old_syntax)

def _nfsopt_normalize_version(nfsopt, old_syntax):
    otheropts = []
    major = None
    minor = None

    for n in nfsopt.split(','):
        if n.startswith('vers='):
            val = n.split('=', 1)[1]

            if val.find('.') >= 0:
                val = val.split('.', 1)
                major = int(val[0])
                minor = int(val[1])
            else:
                major = int(val)

        elif n.startswith('minorversion='):
            val = n.split('=', 1)[1]
            minor = int(val)

        elif n.startswith('v'):
            val = n[1:]

            if val and val[0].isdigit() and val.find('.') >= 0:
                val = val.split('.', 1)
                major = int(val[0])
                minor = int(val[1])

            elif val.isdigit():
                major = int(val)

            else:
                otheropts.append(n)

        else:
            otheropts.append(n)

    vers = None
    if major == 2:
        if old_syntax:
            vers = 'vers=2'
        else:
            vers = 'v2'

    elif major == 3:
        if old_syntax:
            vers = 'vers=3'
        else:
            vers = 'v3'

    elif major == 4 and minor in (None, 0):
        if old_syntax:
            vers = 'vers=4'
        else:
            vers = 'v4'

    elif major == 4 and minor == 1:
        if old_syntax:
            vers = 'vers=4,minorversion=1'
        else:
            vers = 'v4.1'

    if vers:
        new = []
        otheropts.sort()
        new.append(vers)
        new.extend(otheropts)
        nfsopt = ','.join(new)

    else:
        warn("can't normalize version for %r" % (self.__attrs['mountopts']))

    return nfsopt

def nfsopt_split(nfsopt):
    split_tags = nfsopt.split(',')

    assert split_tags

    version_tags = []
    server_tags = []
    client_tags = []

    for tag in split_tags:
        tagtype = TAG_DESC.get(tag, TAGTYPE_USER)

        if tagtype == TAGTYPE_MOUNTOPT_VERS:
            version_tags.append(tag)
        elif tagtype == TAGTYPE_DETECTED:
            server_tags.append(tag)
        elif tagtype == TAGTYPE_USER:
            client_tags.append(tag)
        else:
            raise NotImplemented

    r = (tuple(version_tags), tuple(server_tags), tuple(client_tags))
    return r

def nfsopt_version(nfsopt):
    """
        return the version portion of the nfsopt as one comma sep'd string
    """
    return ','.join(nfsopt_split(nfsopt)[0])

def nfsopt_to_mountopts_tags(nfsopt):
    mt, st, ct = nfsopt_split(nfsopt)

    mountopts = ','.join(mt)

    tags = []
    tags.extend(st)
    tags.extend(ct)

    return (mountopts, tags)

